@@ -23,6 +23,7 @@ from rlgraph.agents import Agent
 from rlgraph.components import Synchronizable, Memory, PrioritizedReplay, DictMerger, \
     ContainerSplitter, DQFDLossFunction
 from rlgraph.spaces import FloatBox, BoolBox
+from rlgraph.utils import RLGraphError
 from rlgraph.utils.decorators import rlgraph_api
 from rlgraph.utils.util import strip_list
 
@@ -58,15 +59,23 @@ class DQFDAgent(Agent):
                 Default: False.
         """
         # Fix action-adapter before passing it to the super constructor.
-        action_adapter_spec = kwargs.pop("action_adapter_spec", dict())
-        # Use a DuelingActionAdapter (instead of a basic ActionAdapter) if option is set.
+        policy_spec = kwargs.pop("policy_spec", dict())
+        # Use a DuelingPolicy (instead of a basic Policy) if option is set.
         if dueling_q is True:
-            action_adapter_spec["type"] = "dueling-action-adapter"
-            assert "units_state_value_stream" in action_adapter_spec
-            assert "units_advantage_stream" in action_adapter_spec
+            policy_spec["type"] = "dueling-policy"
+            # Give us some default state-value nodes.
+            if "units_state_value_stream" not in policy_spec:
+                policy_spec["units_state_value_stream"] = 128
         super(DQFDAgent, self).__init__(
-            action_adapter_spec=action_adapter_spec, name=kwargs.pop("name", "dqn-agent"), **kwargs
+            policy_spec=policy_spec, name=kwargs.pop("name", "dqfd-agent"), **kwargs
         )
+        # Assert that the synch interval is a multiple of the update_interval.
+        if self.update_spec["sync_interval"] / self.update_spec["update_interval"] != \
+                self.update_spec["sync_interval"] // self.update_spec["update_interval"]:
+            raise RLGraphError(
+                "ERROR: sync_interval ({}) must be multiple of update_interval "
+                "({})!".format(self.update_spec["sync_interval"], self.update_spec["update_interval"])
+            )
 
         self.double_q = double_q
         self.dueling_q = dueling_q
@@ -87,7 +96,7 @@ class DQFDAgent(Agent):
 
         self.input_spaces.update(dict(
             actions=self.action_space.with_batch_rank(),
-            weights="variables:policy",
+            weights="variables:{}".format(self.policy.scope),
             time_step=int,
             use_exploration=bool,
             demo_batch_size=int,
@@ -114,7 +123,6 @@ class DQFDAgent(Agent):
 
         # Copy our Policy (target-net), make target-net synchronizable.
         self.target_policy = self.policy.copy(scope="target-policy", trainable=False)
-        self.target_policy.add_components(Synchronizable(), expose_apis="sync")
         # Number of steps since the last target-net synching from the main policy.
         self.steps_since_target_net_sync = 0
 
@@ -132,7 +140,7 @@ class DQFDAgent(Agent):
         self.root_component.add_components(*sub_components)
 
         # Define the Agent's (root-Component's) API.
-        self.define_graph_api("policy", "preprocessor-stack", self.optimizer.scope, *sub_components)
+        self.define_graph_api(self.policy.scope,  self.preprocessor.scope, self.optimizer.scope, *sub_components)
 
         if self.auto_build:
             self._build_graph([self.root_component], self.input_spaces, optimizer=self.optimizer,
@@ -155,7 +163,7 @@ class DQFDAgent(Agent):
         # Act from preprocessed states.
         @rlgraph_api(component=self.root_component)
         def action_from_preprocessed_state(self, preprocessed_states, time_step=0, use_exploration=True):
-            sample_deterministic = policy.get_max_likelihood_action(preprocessed_states)
+            sample_deterministic = policy.get_deterministic_action(preprocessed_states)
             actions = exploration.get_action(sample_deterministic["action"], time_step, use_exploration)
             return preprocessed_states, actions
 
