@@ -17,6 +17,7 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
+from rlgraph.components.common.batch_apply import BatchApply
 from rlgraph.components.common.container_splitter import ContainerSplitter
 from rlgraph.components.common.dict_merger import DictMerger
 from rlgraph.components.common.synchronizable import Synchronizable
@@ -50,7 +51,7 @@ NoiseComponent.__lookup_classes__ = dict(
 NoiseComponent.__default_constructor__ = GaussianNoise
 
 
-__all__ = ["ContainerSplitter", "DictMerger",
+__all__ = ["BatchApply", "ContainerSplitter", "DictMerger",
            "Synchronizable", "RepeaterStack", "Slice",
            "DecayComponent", "ConstantDecay", "LinearDecay", "PolynomialDecay", "ExponentialDecay",
            "NoiseComponent", "ConstantNoise", "GaussianNoise", "OrnsteinUhlenbeckNoise",
