@@ -26,8 +26,8 @@ import numpy as np
 import re
 
 from rlgraph import get_backend
-from rlgraph.utils.decorators import rlgraph_api, component_api_registry, component_graph_fn_registry, define_api_method, \
-    define_graph_fn
+from rlgraph.utils.decorators import rlgraph_api, component_api_registry, component_graph_fn_registry,\
+    define_api_method, define_graph_fn
 from rlgraph.utils.rlgraph_errors import RLGraphError
 from rlgraph.utils.specifiable import Specifiable
 from rlgraph.utils.ops import DataOpDict, FLAT_TUPLE_OPEN, FLAT_TUPLE_CLOSE
@@ -40,6 +40,7 @@ elif get_backend() == "tf-eager":
     import tensorflow.contrib.eager as eager
 elif get_backend() == "pytorch":
     from rlgraph.utils import PyTorchVariable
+    import torch
 
 
 class Component(Specifiable):
@@ -75,9 +76,6 @@ class Component(Specifiable):
             device (str): Device this component will be assigned to. If None, defaults to CPU.
             trainable (Optional[bool]): Whether to make the variables of this Component always trainable or not.
                 Use None for no specific preference.
-            #global_component (bool): In distributed mode, this flag indicates if the component is part of the
-            #    shared global model or local to the worker. Defaults to False and will be ignored if set to
-            #    True in non-distributed mode.
 
             # TODO: remove when we have numpy-based Components (then we can do test calls to infer everything automatically)
             graph_fn_num_outputs (dict): A dict specifying which graph_fns have how many return values.
@@ -85,8 +83,12 @@ class Component(Specifiable):
                 utility function cannot determine the actual number of returned values.
 
             switched_off_apis (Optional[Set[str]]): Set of API-method names that should NOT be build for this Component.
+            
             backend (str): The custom backend that this Component obliges to. None to use the RLGraph global backend.
                 Default: None.
+            
+            space_agnostic (bool): Whether this component does not care about input spaces (e.g. if it does not
+                create any space-dependent variables). Default: False.
         """
         super(Component, self).__init__()
 
@@ -110,6 +112,7 @@ class Component(Specifiable):
         self.graph_fn_num_outputs = kwargs.pop("graph_fn_num_outputs", dict())
         self.switched_off_apis = kwargs.pop("switched_off_apis", set())
         self.backend = kwargs.pop("backend", None)
+        self.space_agnostic = kwargs.pop("space_agnostic", None)
 
         assert not kwargs, "ERROR: kwargs ({}) still contains items!".format(kwargs)
 
@@ -150,8 +153,7 @@ class Component(Specifiable):
         # Set of op-records that are constant and thus can be processed right away at the beginning of the build
         # procedure.
         self.constant_op_records = set()
-        # Whether we know already all our in-Sockets' Spaces.
-        # Only then can we create our variables. Model will do this.
+        # Whether we know already all our API-methods' call args' spaces.
         self.input_complete = False
         # Whether all our sub-Components are input-complete. Only after that point, we can run our _variables graph_fn.
         self.variable_complete = False
@@ -229,12 +231,20 @@ class Component(Specifiable):
         Input-completeness is reached (only once and then it stays that way) if all API-methods of this component
         (whose `must_be_complete` field is not set to False) have all their input Spaces defined.
 
+        Also, if no variables are generated by this Component anyway, the component is input-complete.
+
         Returns:
-            bool: Whether this Component is input_complete or not.
+            bool: Whether this Component is input-complete or not.
         """
         assert self.input_complete is False
 
         self.input_complete = True
+
+        # Short-circuit-set to True, if no variables are generated by this Component anyway.
+        if re.search(r'\spass\n$', inspect.getsource(self.create_variables)) and \
+                re.search(r'\spass\n$', inspect.getsource(self.check_input_spaces)):
+            return True
+
         # Loop through all API methods.
         for method_name, api_method_rec in self.api_methods.items():
             # This API method doesn't have to be completed, ignore and don't add it to space_dict.
@@ -285,8 +295,11 @@ class Component(Specifiable):
         elif self.input_complete is False:
             return False
 
-        # Simply check all sub-Components for input-completeness.
-        self.variable_complete = all(sc.input_complete for sc in self.sub_components.values())
+        # Simply check all direct sub-Components for variable-completeness.
+        for direct_child in self.sub_components.values():
+            if not direct_child.check_variable_completeness():
+                return False
+        self.variable_complete = True
         return self.variable_complete
 
     def when_input_complete(self, input_spaces=None, action_space=None, device=None, summary_regexp=None):
@@ -542,6 +555,7 @@ class Component(Specifiable):
         Private variable from space helper, see 'get_variable' for API.
         """
         # Variables should be returned in a flattened OrderedDict.
+        # TODO can we hide this tf.variable_scope somewhere?
         if get_backend() == "tf":
             if self.reuse_variable_scope is not None:
                 with tf.variable_scope(name_or_scope=self.reuse_variable_scope, reuse=True):
@@ -575,6 +589,20 @@ class Component(Specifiable):
                         local=local, use_resource=use_resource
                     )
 
+        elif get_backend() == "pytorch":
+            if flatten:
+                return from_space.flatten(mapping=lambda key_, primitive: primitive.get_variable(
+                    name=name + key_, add_batch_rank=add_batch_rank, add_time_rank=add_time_rank,
+                    time_major=time_major, trainable=trainable, initializer=initializer,
+                    is_python=True, local=local, use_resource=use_resource
+                ))
+            # Normal, nested Variables from a Space (container or primitive).
+            else:
+                return from_space.get_variable(
+                    name=name, add_batch_rank=add_batch_rank, trainable=trainable, initializer=initializer,
+                    is_python=True, local=local, use_resource=use_resource
+                )
+
     def get_variables(self, *names, **kwargs):
         """
         Utility method to get one or more component variable(s) by name(s).
@@ -766,6 +794,10 @@ class Component(Specifiable):
             expose_apis = {expose_apis}
 
         for component in components:
+            # Safety measure: Ignore Nones.
+            if component is None:
+                continue
+
             # Try to create Component from spec.
             if not isinstance(component, Component):
                 component = Component.from_spec(component)
@@ -783,12 +815,6 @@ class Component(Specifiable):
             component.parent_component = self
             self.sub_components[component.name] = component
 
-            # Add own reusable scope to front of sub-components'.
-            if self.reuse_variable_scope is not None:
-                # Propagate reuse_variable_scope down to the added Component's sub-components.
-                component.propagate_sub_component_properties(
-                    properties=dict(reuse_variable_scope=self.reuse_variable_scope)
-                )
             # Fix the sub-component's (and sub-sub-component's etc..) scope(s).
             self.propagate_scope(component)
 
@@ -811,19 +837,30 @@ class Component(Specifiable):
                         # component gets copied.
                         return getattr(self.sub_components[component_name], name_)(*inputs)
 
-    def get_all_sub_components(self, list_=None, level_=0):
+        # Add own reusable scope to front of all sub-components' reusable scope.
+        if self.reuse_variable_scope is not None:
+            # Propagate reuse_variable_scope down to the added Component's sub-components.
+            self.propagate_sub_component_properties(
+                properties=dict(reuse_variable_scope=self.reuse_variable_scope)
+            )
+
+    def get_all_sub_components(self, list_=None, level_=0, exclude_self=False):
         """
-        Returns all sub-Components (including self) sorted by their nesting-level (... grand-children before children
-        before parents).
+        Returns all sub-Components (including self, unless `exclude_self` is True) sorted by their nesting-level
+        (... grand-children before children before parents). If the nesting level is the same, sort alphabetically
+        by the scope (name) of the Components.
 
         Args:
             list\_ (Optional[List[Component]])): A list of already collected components to append to.
             level\_ (int): The slot indicating the Component level depth in `list_` at which we are currently.
+            exclude_self (bool): Whether `self` should be returned as the last sub-Component in the list.
+                Default: True.
 
         Returns:
-            List[Component]: A list with all the sub-components in `self` and `self` itself.
+            List[Component]: A list with all the sub-components in `self` (and `self` itself if `exclude_self` is False).
         """
         return_ = False
+        # This is the final-return recursive call-level.
         if list_ is None:
             list_ = dict()
             return_ = True
@@ -837,6 +874,8 @@ class Component(Specifiable):
             ret = list()
             for l in sorted(list_.keys(), reverse=True):
                 ret.extend(sorted(list_[l], key=lambda c: c.scope))
+            if exclude_self:
+                return ret[:-1]
             return ret
 
     def get_sub_component_by_global_scope(self, scope):
@@ -929,7 +968,7 @@ class Component(Specifiable):
         for sc in sub_component.sub_components.values():
             sub_component.propagate_scope(sc)
 
-    def propagate_sub_component_properties(self, properties, component=None):
+    def propagate_sub_component_properties(self, properties, component=None, recursive_=False):
         """
         Recursively updates properties of component and its sub-components.
 
@@ -938,6 +977,8 @@ class Component(Specifiable):
                 sub-components with.
 
             component (Optional([Component])): Component to recursively update. Uses self if None.
+
+            recursive_ (bool): Whether this is a recursive (sub-Component) call. Default: False.
         """
         if component is None:
             component = self
@@ -946,14 +987,16 @@ class Component(Specifiable):
             # Property is some scope (value is then that scope of the parent component).
             # Have to align it with sub-component's local scope.
             if value and (name == "global_scope" or name == "reuse_variable_scope"):
-                value += (("/" + component.scope) if component.scope else "")
+                # For the parent component, do not add its scope to the shared-scope.
+                if recursive_ is True:
+                    value += (("/" + component.scope) if component.scope else "")
                 properties_scoped[name] = value
                 setattr(component, name, value)
             # Normal property: Set to static given value.
             else:
                 setattr(component, name, value)
         for sc in component.sub_components.values():
-            component.propagate_sub_component_properties(properties_scoped, sc)
+            component.propagate_sub_component_properties(properties_scoped, sc, recursive_=True)
 
     def propagate_variables(self, keys=None):
         """
@@ -1068,10 +1111,11 @@ class Component(Specifiable):
             value (any): The value to use for the assignment.
 
         Returns:
-            Optional[op]: None or the graph operation representing the assginment.
+            Optional[op]: None or the graph operation representing the assignment.
         """
         if get_backend() == "tf":
-            if type(value).__name__ == "Variable":
+            tensor_type = type(value).__name__
+            if tensor_type == "Variable" or tensor_type == "RefVariable":
                 return tf.assign(ref=ref, value=value.read_value())
             else:
                 return tf.assign(ref=ref, value=value)
@@ -1079,7 +1123,7 @@ class Component(Specifiable):
             ref.set_value(value)
 
     @staticmethod
-    def read_variable(variable, indices=None):
+    def read_variable(variable, indices=None, dtype=None):
         """
         Reads a variable.
 
@@ -1098,12 +1142,27 @@ class Component(Specifiable):
             else:
                 return variable
         elif get_backend() == "pytorch":
-            if not isinstance(variable, PyTorchVariable):
-                raise RLGraphError("Variable must be of type PyTorchVariable but is {}.".format(
-                    type(variable)
-                ))
-            else:
+            # PyTorchVariable is used to store torch parameters (e.g. layers).
+            if isinstance(variable, PyTorchVariable):
                 return variable.get_value()
+            # Lists or numpy arrays may be used to store mutable state that does not need
+            # tensor operations.
+            elif isinstance(variable, list) or isinstance(variable, np.ndarray):
+                ret = []
+                for i in indices:
+                    val = variable[i]
+                    # Type checking is necessary because torch.stack only works on same types.
+                    if isinstance(val, torch.Tensor):
+                        if dtype is None:
+                            ret.append(val)
+                        elif dtype == torch.float32:
+                            ret.append(val.float())
+                        elif dtype == torch.int32:
+                            ret.append(val.int())
+                        elif dtype == torch.uint8:
+                            ret.append(val.byte())
+                # Stack list into one Tensor with a btach dim.
+                return torch.stack(ret)
 
     def sub_component_by_name(self, scope_name):
         """
