@@ -118,7 +118,8 @@ class LSTMLayer(NNLayer):
                 use_peephole=self.use_peepholes,
                 cell_clip=self.cell_clip,
                 forget_bias=self.forget_bias,
-                name="lstm-cell"
+                name="lstm-cell",
+                reuse=tf.AUTO_REUSE
                 # TODO: self.trainable needs to be recognized somewhere here.
 
                 # These are all not supported yet for LSTMBlockCell (only for the slower LSTMCell)
@@ -140,7 +141,9 @@ class LSTMLayer(NNLayer):
 
     @rlgraph_api
     def apply(self, inputs, initial_c_and_h_states=None, sequence_length=None):
-        output, last_internal_states = self._graph_fn_apply(inputs, initial_c_and_h_states, sequence_length)
+        output, last_internal_states = self._graph_fn_apply(
+            inputs, initial_c_and_h_states=initial_c_and_h_states, sequence_length=sequence_length
+        )
         return dict(output=output, last_internal_states=last_internal_states)
 
     @graph_fn
@@ -189,7 +192,6 @@ class LSTMLayer(NNLayer):
                 inputs.set_shape([self.static_loop] + inputs.shape.as_list()[1:])
                 #for input_, terminal in zip(tf.unstack(inputs), tf.unstack(terminals)):
                 for input_ in tf.unstack(inputs):
-                    #input_ = inputs[i]
                     # If the episode ended, the core state should be reset before the next.
                     #core_state = nest.map_structure(functools.partial(tf.where, d),
                     #                                initial_core_state, core_state)
