@@ -18,7 +18,7 @@ from __future__ import division
 from __future__ import print_function
 
 from rlgraph import get_backend
-from rlgraph.utils.decorators import rlgraph_api
+from rlgraph.utils.decorators import rlgraph_api, graph_fn
 from rlgraph.components.loss_functions.dqn_loss_function import DQNLossFunction
 from rlgraph.utils.util import get_rank
 
@@ -57,61 +57,47 @@ class DQFDLossFunction(DQNLossFunction):
         return total_loss, loss_per_item
 
     @rlgraph_api
-    def _graph_fn_loss_per_item(self, q_values_s, actions, rewards, terminals,
-                                qt_values_sp, q_values_sp=None, importance_weights=None, apply_demo_loss=False):
+    def loss_per_item(self, q_values_s, actions, rewards, terminals, qt_values_sp, q_values_sp=None,
+                      importance_weights=None, apply_demo_loss=False):
+
+        # Get the targets -> inherited from DQNLossFunction
+        td_targets = self._graph_fn_get_td_targets(rewards, terminals, qt_values_sp, q_values_sp)
+        # Average over container sub-actions.
+        td_targets = self._graph_fn_average_over_container_keys(td_targets)
+
+        # Calculate the loss per item.
+        loss_per_item = self._graph_fn_loss_per_item(td_targets, q_values_s, actions,
+                                                     importance_weights, apply_demo_loss)
+        # Average over container sub-actions.
+        loss_per_item = self._graph_fn_average_over_container_keys(loss_per_item)
+
+        return loss_per_item
+
+    @graph_fn(flatten_ops=True, split_ops=True, add_auto_key_as_first_param=True)
+    def _graph_fn_loss_per_item(self, key, td_targets, q_values_s, actions,
+                                importance_weights=None, apply_demo_loss=False):
         """
         Args:
+            td_targets (SingleDataOp): The already calculated TD-target terms (r + gamma maxa'Qt(s',a')
+                OR for double Q: r + gamma Qt(s',argmaxa'(Q(s',a'))))
             q_values_s (SingleDataOp): The batch of Q-values representing the expected accumulated discounted returns
                 when in s and taking different actions a.
             actions (SingleDataOp): The batch of actions that were actually taken in states s (from a memory).
-            rewards (SingleDataOp): The batch of rewards that we received after having taken a in s (from a memory).
-            terminals (SingleDataOp): The batch of terminal signals that we received after having taken a in s
-                (from a memory).
-            qt_values_sp (SingleDataOp): The batch of Q-values representing the expected accumulated discounted
-                returns (estimated by the target net) when in s' and taking different actions a'.
-            q_values_sp (Optional[SingleDataOp]): If `self.double_q` is True: The batch of Q-values representing the
-                expected accumulated discounted returns (estimated by the (main) policy net) when in s' and taking
-                different actions a'.
             importance_weights (Optional[SingleDataOp]): If 'self.importance_weights' is True: The batch of weights to
                 apply to the losses.
             apply_demo_loss (Optional[SingleDataOp]): If 'apply_demo_loss' is True: The large-margin loss is applied.
                 Should be set to True when updating from demo data, False when updating from online data.
+
         Returns:
             SingleDataOp: The loss values vector (one single value for each batch item).
         """
         if get_backend() == "tf":
-            # Make sure the target policy's outputs are treated as constant when calculating gradients.
-            qt_values_sp = tf.stop_gradient(qt_values_sp)
-
-            if self.double_q:
-                # For double-Q, we no longer use the max(a')Qt(s'a') value.
-                # Instead, the a' used to get the Qt(s'a') is given by argmax(a') Q(s',a') <- Q=q-net, not target net!
-                a_primes = tf.argmax(input=q_values_sp, axis=-1)
-
-                # Now lookup Q(s'a') with the calculated a'.
-                one_hot = tf.one_hot(indices=a_primes, depth=self.action_space.num_categories)
-                qt_sp_ap_values = tf.reduce_sum(input_tensor=(qt_values_sp * one_hot), axis=-1)
-            else:
-                # Qt(s',a') -> Use the max(a') value (from the target network).
-                qt_sp_ap_values = tf.reduce_max(input_tensor=qt_values_sp, axis=-1)
-
-            # Make sure the rewards vector (batch) is broadcast correctly.
-            for _ in range(get_rank(qt_sp_ap_values) - 1):
-                rewards = tf.expand_dims(rewards, axis=1)
-
-            # Ignore Q(s'a') values if s' is a terminal state. Instead use 0.0 as the state-action value for s'a'.
-            # Note that in that case, the next_state (s') is not the correct next state and should be disregarded.
-            # See Chapter 3.4 in "RL - An Introduction" (2017 draft) by A. Barto and R. Sutton for a detailed analysis.
-            qt_sp_ap_values = tf.where(
-                condition=terminals, x=tf.zeros_like(qt_sp_ap_values), y=qt_sp_ap_values
-            )
-
             # Q(s,a) -> Use the Q-value of the action actually taken before.
-            one_hot = tf.one_hot(indices=actions, depth=self.action_space.num_categories)
+            one_hot = tf.one_hot(indices=actions, depth=self.flat_action_space[key].num_categories)
             q_s_a_values = tf.reduce_sum(input_tensor=(q_values_s * one_hot), axis=-1)
 
-            # Calculate the TD-delta (target - current estimate).
-            td_delta = (rewards + (self.discount ** self.n_step) * qt_sp_ap_values) - q_s_a_values
+            # Calculate the TD-delta (targets - current estimate).
+            td_delta = td_targets - q_s_a_values
 
             # Calculate the demo-loss.
             #  J_E(Q) = max_a([Q(s, a_taken) + l(s, a_expert, a_taken)] - Q(s, a_expert)
