@@ -22,7 +22,7 @@ from rlgraph.components.neural_networks.dict_preprocessor_stack import DictPrepr
 from rlgraph.components.neural_networks.preprocessor_stack import PreprocessorStack
 
 from rlgraph.components.neural_networks.neural_network import NeuralNetwork
-from rlgraph.components.neural_networks.policy import Policy
+from rlgraph.components.neural_networks.value_function import ValueFunction
 
 # The Stacks.
 Stack.__lookup_classes__ = dict(
@@ -30,6 +30,6 @@ Stack.__lookup_classes__ = dict(
     preprocessorstack=PreprocessorStack
 )
 
-__all__ = ["NeuralNetwork", "Policy"] + \
+__all__ = ["NeuralNetwork", "ValueFunction"] + \
           ["Stack"] + \
           list(set(map(lambda x: x.__name__, Stack.__lookup_classes__.values())))
