@@ -18,7 +18,6 @@ from __future__ import division
 from __future__ import print_function
 
 from copy import deepcopy
-
 import numpy as np
 from rlgraph.components import PreprocessorStack
 from six.moves import xrange as range_
@@ -37,9 +36,13 @@ class SingleThreadedWorker(Worker):
         self.logger.info("Initialized single-threaded executor with {} environments '{}' and Agent '{}'".format(
             self.num_environments, self.vector_env.get_env(), self.agent
         ))
+
+        # Switch off worker preprocessing if nothing to do anyway.
+        if preprocessing_spec is None or preprocessing_spec == []:
+            worker_executes_preprocessing = False
+
         self.worker_executes_preprocessing = worker_executes_preprocessing
         if self.worker_executes_preprocessing:
-            assert preprocessing_spec is not None
             self.preprocessors = {}
             self.state_is_preprocessed = {}
             for env_id in self.env_ids:
@@ -173,6 +176,7 @@ class SingleThreadedWorker(Worker):
         episode_terminals = self.episode_terminals
         if reset is True:
             self.env_frames = 0
+            self.episodes_since_update = 0
             self.finished_episode_rewards = [[] for _ in range_(self.num_environments)]
             self.finished_episode_durations = [[] for _ in range_(self.num_environments)]
             self.finished_episode_timesteps = [[] for _ in range_(self.num_environments)]
@@ -221,8 +225,25 @@ class SingleThreadedWorker(Worker):
             # Accumulate the reward over n env-steps (equals one action pick). n=self.frameskip.
             env_rewards = [0 for _ in range_(self.num_environments)]
             next_states = None
+            # For container action spaces, we have to treat each key as an array with batch-rank at index 0.
+            # The action-dict is then translated into a list of dicts where each dict contains the original data
+            # but without the batch-rank.
+            # E.g. {'A': array([0, 1]), 'B': array([2, 3])} -> [{'A': 0, 'B': 2}, {'A': 1, 'B': 3}]
+            if self.agent.flat_action_space is not None:
+                some_key = next(iter(actions))
+                assert isinstance(actions, dict) and isinstance(actions[some_key], np.ndarray),\
+                    "ERROR: Cannot flip container-action batch with dict keys if returned value is not a dict OR " \
+                    "values of returned value are not np.ndarrays!"
+                # TODO: What if actions come as nested dicts (more than one level deep)?
+                env_actions = [{key: value[i] for key, value in actions.items()} for i in range(len(actions[some_key]))]
+            # No flipping necessary.
+            else:
+                env_actions = actions
+                if self.num_environments == 1 and env_actions.shape == ():
+                    env_actions = [env_actions]
+
             for _ in range_(frameskip):
-                next_states, step_rewards, episode_terminals, infos = self.vector_env.step(actions=actions)
+                next_states, step_rewards, episode_terminals, infos = self.vector_env.step(actions=env_actions)
 
                 self.env_frames += self.num_environments
                 for i, step_reward in enumerate(step_rewards):
@@ -231,8 +252,8 @@ class SingleThreadedWorker(Worker):
                     break
 
             # Only render once per action.
-            if self.render:
-                self.vector_env.environments[0].render()
+            #if self.render:
+            #    self.vector_env.environments[0].render()
 
             for i, env_id in enumerate(self.env_ids):
                 self.episode_returns[i] += env_rewards[i]
@@ -245,11 +266,18 @@ class SingleThreadedWorker(Worker):
                 # Do accounting for finished episodes.
                 if episode_terminals[i]:
                     episodes_executed += 1
+                    self.episodes_since_update += 1
+                    episode_duration = time.perf_counter() - self.episode_starts[i]
                     self.finished_episode_rewards[i].append(self.episode_returns[i])
-                    self.finished_episode_durations[i].append(time.perf_counter() - self.episode_starts[i])
+                    self.finished_episode_durations[i].append(episode_duration)
                     self.finished_episode_timesteps[i].append(self.episode_timesteps[i])
-                    self.logger.debug("Finished episode: reward={}, actions={}, duration={}s.".format(
-                        self.episode_returns[i], self.episode_timesteps[i], self.finished_episode_durations[i][-1]))
+
+                    self.log_finished_episode(
+                        reward=self.episode_returns[i],
+                        duration=episode_duration,
+                        timesteps=self.episode_timesteps[i],
+                        env_num=i
+                    )
 
                     # Reset this environment and its preprocecssor stack.
                     env_states[i] = self.vector_env.reset(i)
@@ -269,12 +297,12 @@ class SingleThreadedWorker(Worker):
                     env_states[i] = next_states[i]
 
                 if self.worker_executes_preprocessing and self.preprocessors[env_id] is not None:
-                    next_state = self.agent.state_space.force_batch(env_states[i])
-                    next_states[i] = np.array(self.preprocessors[env_id].preprocess(next_state))
+                    #next_state = self.agent.state_space.force_batch(env_states[i])
+                    next_states[i] = np.array(self.preprocessors[env_id].preprocess(env_states[i]))  # next_state
                 # TODO: If worker does not execute preprocessing, next state is not preprocessed here.
                 # Observe per environment.
                 self.agent.observe(
-                    preprocessed_states=preprocessed_states[i], actions=actions[i], internals=[],
+                    preprocessed_states=preprocessed_states[i], actions=env_actions[i], internals=[],
                     rewards=env_rewards[i], next_states=next_states[i],
                     terminals=episode_terminals[i], env_id=self.env_ids[i]
                 )
