@@ -30,7 +30,7 @@ class Worker(Specifiable):
     Generic worker to locally interact with simulator environments.
     """
     def __init__(self, agent, env_spec=None, num_envs=1, frameskip=1, render=False,
-                 worker_executes_exploration=True, exploration_epsilon=0.1):
+                 worker_executes_exploration=True, exploration_epsilon=0.1, episode_finish_callback=None):
         """
         Args:
             agent (Agent): Agent to execute environment on.
@@ -63,10 +63,16 @@ class Worker(Specifiable):
         self.update_interval = None
         self.update_steps = None
         self.sync_interval = None
+        self.episodes_since_update = 0
+
+        # Default val or None?
+        self.update_mode = "time_steps"
 
         self.worker_executes_exploration = worker_executes_exploration
         self.exploration_epsilon = exploration_epsilon
 
+        self.episode_finish_callback = episode_finish_callback
+
     def execute_timesteps(self, num_timesteps, max_timesteps_per_episode=0, update_spec=None, use_exploration=True,
                           frameskip=1, reset=True):
         """
@@ -174,20 +180,36 @@ class Worker(Specifiable):
         if self.updating:
             # Are we allowed to update?
             if self.agent.timesteps > self.steps_before_update and \
-                    (self.agent.observe_spec["buffer_enabled"] is False or  # no update before some data in buffer
-                     self.agent.timesteps >= self.agent.observe_spec["buffer_size"]) and \
-                    self.agent.timesteps % self.update_interval == 0:  # update frequency check
-                loss = 0
-                for _ in range_(self.update_steps):
-                    ret = self.agent.update()
-                    if isinstance(ret, tuple):
-                        loss += ret[0]
-                    else:
-                        loss += ret
-                return loss
-
+                    (self.agent.observe_spec["buffer_enabled"] is False or  # No update before some data in buffer
+                     self.agent.timesteps >= self.agent.observe_spec["buffer_size"]):
+                # Updating according to one update mode:
+                if self.update_mode == "time_steps" and self.agent.timesteps % self.update_interval == 0:
+                    return self.execute_update()
+                elif self.update_mode == "episodes" and self.episodes_since_update == self.update_interval:
+                    # Do not do modulo here - this would be called every step in one episode otherwise.
+                    loss = self.execute_update()
+                    self.episodes_since_update = 0
+                    return loss
         return None
 
+    def execute_update(self):
+        loss = 0
+        for _ in range_(self.update_steps):
+            ret = self.agent.update()
+            # TEST:
+            #print("current Q-table:\n{}".format(self.agent.last_q_table))
+            # Find q-values for state (0, 1, 0, 1)
+            #for i, state in enumerate(self.agent.last_q_table["states"]):
+            #    if tuple(state) == (0.0, 1.0, 0.0, 1.0):
+            #        print("Q-values for 2nd state: {}".format(self.agent.last_q_table["q_values"][i]))
+            #        break
+            # END: TEST
+            if isinstance(ret, tuple):
+                loss += ret[0]
+            else:
+                loss += ret
+        return loss
+
     def set_update_schedule(self, update_schedule=None):
         """
         Sets this worker's update schedule. By default, a worker is not updating but only acting
@@ -206,6 +228,9 @@ class Worker(Specifiable):
             self.update_steps = update_schedule['update_steps']
             self.sync_interval = update_schedule['sync_interval']
 
+            # Interpret update interval as n time-steps or n episodes.
+            self.update_mode = update_schedule.get("update_mode", "time_steps")
+
     def get_action(self, states, use_exploration, apply_preprocessing, extra_returns):
         if self.worker_executes_exploration:
             # Only once for all actions otherwise we would have to call a session anyway.
@@ -227,4 +252,9 @@ class Worker(Specifiable):
             return self.agent.get_action(states=states, use_exploration=use_exploration,
                                          apply_preprocessing=apply_preprocessing, extra_returns=extra_returns)
 
+    def log_finished_episode(self, reward, duration, timesteps, env_num=0):
+        self.logger.debug("Finished episode: reward={}, actions={}, duration={}s.".format(
+            reward, duration, timesteps))
 
+        if self.episode_finish_callback:
+            self.episode_finish_callback(reward=reward, duration=duration, timesteps=timesteps, env_num=env_num)
