@@ -22,8 +22,11 @@ import re
 import time
 
 import inspect
+from collections import OrderedDict
+
 from rlgraph import get_backend
-from rlgraph.utils.rlgraph_errors import RLGraphError
+from rlgraph.utils.execution_util import define_by_run_flatten, define_by_run_split_args, define_by_run_unflatten
+from rlgraph.utils.rlgraph_errors import RLGraphError, RLGraphBuildError
 from rlgraph.utils.specifiable import Specifiable
 
 from rlgraph.components.component import Component
@@ -31,14 +34,15 @@ from rlgraph.spaces import Space, Dict
 from rlgraph.spaces.space_utils import get_space_from_op, check_space_equivalence
 from rlgraph.utils.input_parsing import parse_summary_spec
 from rlgraph.utils.util import force_list, force_tuple, get_shape
-from rlgraph.utils.ops import DataOpTuple, is_constant
+from rlgraph.utils.ops import is_constant, ContainerDataOp, flatten_op
 from rlgraph.utils.op_records import FlattenedDataOp, DataOpRecord, DataOpRecordColumnIntoGraphFn, \
     DataOpRecordColumnIntoAPIMethod, DataOpRecordColumnFromGraphFn, get_call_param_name
-from rlgraph.utils.component_printout import component_print_out
 
 if get_backend() == "tf":
     import tensorflow as tf
     from rlgraph.utils.tf_util import pin_global_variables
+elif get_backend() == "pytorch":
+    import torch
 
 
 class GraphBuilder(Specifiable):
@@ -47,7 +51,7 @@ class GraphBuilder(Specifiable):
     components, sockets and connections and creating the underlying computation
     graph.
     """
-    def __init__(self, name="model", action_space=None, summary_spec=None):
+    def __init__(self, name="model", action_space=None, summary_spec=None, max_build_iterations=3000):
         """
         Args:
             name (str): The name of this GraphBuilder and of the meta-graph's root-component.
@@ -63,6 +67,9 @@ class GraphBuilder(Specifiable):
         self.name = name
         self.action_space = action_space
         self.summary_spec = parse_summary_spec(summary_spec)
+        self.max_build_iterations = max_build_iterations
+        # Set of components that have been analyzed on why they remain input-incomplete.
+        self.investigated_input_incomplete_components = set()
 
         # All components assigned to each device, for debugging and analysis.
         self.device_component_assignments = dict()
@@ -117,8 +124,8 @@ class GraphBuilder(Specifiable):
             device_map (Optional[Dict]): Dict of Component names mapped to device names to place the Component's ops.
             build_options (Optional[Dict]): Dict of build options, e.g. default device handling for TF.
         """
-        # No build options.
-        if build_options is None:
+        # No device context options.
+        if build_options is None or "build_device_context" not in build_options:
             return self.build_graph(meta_graph, input_spaces, available_devices,
                                     device_strategy, default_device, device_map)
         else:
@@ -163,8 +170,8 @@ class GraphBuilder(Specifiable):
         time_start = time.perf_counter()
         assert meta_graph.build_status, "ERROR: Meta graph must be built to build backend graph."
         self.root_component = meta_graph.root_component
-        self.graph_call_times = list()
-        self.var_call_times = list()
+        self.graph_call_times = []
+        self.var_call_times = []
         self.api = meta_graph.api
         self.num_meta_ops = meta_graph.num_ops
 
@@ -175,7 +182,7 @@ class GraphBuilder(Specifiable):
         self.available_devices = available_devices
         self.device_strategy = device_strategy
         self.default_device = default_device
-        self.device_map = device_map or dict()
+        self.device_map = device_map or {}
 
         # Create the first actual ops based on the input-spaces.
         # Some ops can only be created later when variable-based-Spaces are known (op_records_to_process_later).
@@ -183,10 +190,13 @@ class GraphBuilder(Specifiable):
 
         # Collect all components and add those op-recs to the set that are constant.
         components = self.root_component.get_all_sub_components()
+        # Point to this GraphBuilder object.
+        # Add those op-recs to the set that are constant.
+        for component in components:
+            component.graph_builder = self
+        # Try to build if already input complete.
         for component in components:
-            component.graph_builder = self  # point to us.
             self.op_records_to_process.update(component.constant_op_records)
-            # Check whether the Component is input-complete (and build already if it is).
             self.build_component_when_input_complete(component)
 
         op_records_list = sorted(self.op_records_to_process, key=lambda rec: rec.id)
@@ -202,6 +212,10 @@ class GraphBuilder(Specifiable):
 
         self.num_trainable_parameters = self.count_trainable_parameters()
         self.logger.info("Number of trainable parameters: {}".format(self.num_trainable_parameters))
+
+        # Sanity check the build.
+        self.sanity_check_build()
+
         # The build here is the actual build overhead, so build time minus the tensorflow calls and variable
         # creations which would have to happen either way.
         build_overhead = time_build - sum(self.graph_call_times) - sum(self.var_call_times)
@@ -228,7 +242,7 @@ class GraphBuilder(Specifiable):
         """
         for api_method_name, (in_op_records, _) in sorted(self.api.items()):
             api_method_rec = self.root_component.api_methods[api_method_name]
-            spaces = list()
+            spaces = []
             for param_name in api_method_rec.input_names:
                 if self.root_component.api_method_inputs[param_name] == "flex":
                     if input_spaces is not None and param_name in input_spaces:
@@ -251,7 +265,7 @@ class GraphBuilder(Specifiable):
                         spaces.append(input_spaces[param_name])
             assert len(spaces) == len(in_op_records)
 
-            # Create the placeholder and store it in the given DataOpRecords.
+            # Create the placeholders and store them in the given DataOpRecords.
             for i, space in enumerate(spaces):
                 # Space is dependent on the variables of some sub-component (wait with the construction of the
                 # placeholder until after the build).
@@ -321,10 +335,8 @@ class GraphBuilder(Specifiable):
                 if component.parent_component is not None:
                     self.build_component_when_input_complete(component.parent_component)
 
-        # Check variable-completeness and actually call the _variable graph_fn if the component just became
-        # "variable-complete".
-        if component.input_complete is True and component.variable_complete is False and \
-                component.check_variable_completeness():
+        # Check variable-completeness and actually call the _variable graph_fn if not already done so.
+        if component.input_complete is True and component.check_variable_completeness():
             # The graph_fn _variables has some in-op-columns that need to be run through the function.
             if "_graph_fn__variables" in component.graph_fns:
                 graph_fn_rec = component.graph_fns["_graph_fn__variables"]
@@ -333,10 +345,13 @@ class GraphBuilder(Specifiable):
                 for i, in_op_col in enumerate(graph_fn_rec.in_op_columns):
                     if in_op_col.already_sent is False:
                         self.run_through_graph_fn_with_device_and_scope(in_op_col)
-                        # Keep working with the generated output ops.
+                        # If graph_fn_rec doesn't know about the out-op-col yet, add it.
+                        if len(graph_fn_rec.out_op_columns) <= i:
+                            assert len(graph_fn_rec.out_op_columns) == i  # make sure, it's really just one col missing
+                            graph_fn_rec.out_op_columns.append(in_op_col.out_graph_fn_column)
                         self.op_records_to_process.update(graph_fn_rec.out_op_columns[i].op_records)
 
-    def run_through_graph_fn_with_device_and_scope(self, op_rec_column, create_new_out_column=False):
+    def run_through_graph_fn_with_device_and_scope(self, op_rec_column, create_new_out_column=None):
         """
         Runs through a graph_fn with the given ops and thereby assigns a device (Component's device or GraphBuilder's
         default) to the ops generated by a graph_fn.
@@ -492,7 +507,7 @@ class GraphBuilder(Specifiable):
                 loop_counter += 1
         return subgraph_container
 
-    def run_through_graph_fn(self, op_rec_column, create_new_out_column=False):
+    def run_through_graph_fn(self, op_rec_column, create_new_out_column=None):
         """
         Pushes all ops in the column through the respective graph_fn (graph_fn-spec and call-options are part of
         the column).
@@ -503,10 +518,10 @@ class GraphBuilder(Specifiable):
         Args:
             op_rec_column (DataOpRecordColumnIntoGraphFn): The column of DataOpRecords to be fed through the
                 graph_fn.
-            create_new_out_column (bool): Whether to produce the out op-record column (or use the one already in
-                the meta-graph). If True and the `op_rec_column` already links to an out op-rec column, raises
-                an error.
-                Default: False.
+            create_new_out_column (Optional[bool]): If given, whether to produce the out op-record column
+                (or use the one already in the meta-graph). If True and the `op_rec_column` already links to an out
+                op-rec column, raises an error.
+                Default: None, meaning only create a new column if one dies not exist.
 
         Returns:
             DataOpRecordColumnFromGraphFn: The op-record column coming out of the graph_fn. This column may have
@@ -518,6 +533,7 @@ class GraphBuilder(Specifiable):
         assert all(op is not None for op in args)  # just make sure
 
         call_time = None
+        is_build_time = self.phase == "building"
 
         # Build the ops from this input-combination.
         # Flatten input items.
@@ -528,16 +544,16 @@ class GraphBuilder(Specifiable):
                 split_args_and_kwargs = op_rec_column.split_flattened_input_ops(*flattened_args, **flattened_kwargs)
                 # There is some splitting to do. Call graph_fn many times (one for each split).
                 if isinstance(split_args_and_kwargs, FlattenedDataOp):
-                    ops = dict()
+                    ops = {}
                     num_return_values = -1
                     for key, params in split_args_and_kwargs.items():
                         params_args = [p for p in params if not isinstance(p, tuple)]
                         params_kwargs = {p[0]: p[1] for p in params if isinstance(p, tuple)}
-                        if create_new_out_column is False:
+                        if is_build_time:
                             call_time = time.perf_counter()
                         ops[key] = force_tuple(op_rec_column.graph_fn(op_rec_column.component,
                                                                       *params_args, **params_kwargs))
-                        if create_new_out_column is False:
+                        if is_build_time:
                             self.graph_call_times.append(time.perf_counter() - call_time)
                         if num_return_values >= 0 and num_return_values != len(ops[key]):
                             raise RLGraphError(
@@ -546,7 +562,7 @@ class GraphBuilder(Specifiable):
                             )
                         num_return_values = len(ops[key])
                     # Un-split the results dict into a tuple of `num_return_values` slots.
-                    un_split_ops = list()
+                    un_split_ops = []
                     for i in range(num_return_values):
                         dict_with_singles = FlattenedDataOp()
                         for key in split_args_and_kwargs.keys():
@@ -557,23 +573,23 @@ class GraphBuilder(Specifiable):
                 # No splitting to do: Pass everything as-is.
                 else:
                     split_args, split_kwargs = split_args_and_kwargs[0], split_args_and_kwargs[1]
-                    if create_new_out_column is False:
+                    if is_build_time:
                         call_time = time.perf_counter()
                     ops = op_rec_column.graph_fn(op_rec_column.component, *split_args, **split_kwargs)
-                    if create_new_out_column is False:
+                    if is_build_time:
                         self.graph_call_times.append(time.perf_counter() - call_time)
             else:
-                if create_new_out_column is False:
+                if is_build_time:
                     call_time = time.perf_counter()
                 ops = op_rec_column.graph_fn(op_rec_column.component, *flattened_args, **flattened_kwargs)
-                if create_new_out_column is False:
+                if is_build_time:
                     self.graph_call_times.append(time.perf_counter() - call_time)
         # Just pass in everything as-is.
         else:
-            if create_new_out_column is False:
+            if is_build_time:
                 call_time = time.perf_counter()
             ops = op_rec_column.graph_fn(op_rec_column.component, *args, **kwargs)
-            if create_new_out_column is False:
+            if is_build_time:
                 self.graph_call_times.append(time.perf_counter() - call_time)
 
         # Make sure everything coming from a computation is always a tuple (for out-Socket indexing).
@@ -584,14 +600,23 @@ class GraphBuilder(Specifiable):
         ops = op_rec_column.unflatten_output_ops(*ops)
 
         # Should we create a new out op-rec column?
-        if create_new_out_column is True:
+        if create_new_out_column is not False:
             # Assert that we don't have an out column already (wouldn't make sense).
-            assert op_rec_column.out_graph_fn_column is None
-            out_graph_fn_column = DataOpRecordColumnFromGraphFn(
-                len(ops), component=op_rec_column.component, graph_fn_name=op_rec_column.graph_fn.__name__,
-                in_graph_fn_column=op_rec_column
-            )
+            if create_new_out_column is True and op_rec_column.out_graph_fn_column is not None:
+                raise RLGraphError(
+                    "New DataOpRecordColumnFromGraphFn requested, but one already exists in in-column "
+                    "{}!".format(op_rec_column)
+                )
+            if op_rec_column.out_graph_fn_column is None:
+                out_graph_fn_column = DataOpRecordColumnFromGraphFn(
+                    len(ops), component=op_rec_column.component, graph_fn_name=op_rec_column.graph_fn.__name__,
+                    in_graph_fn_column=op_rec_column
+                )
+            else:
+                out_graph_fn_column = op_rec_column.out_graph_fn_column
         else:
+            assert op_rec_column.out_graph_fn_column is not None,\
+                "ERROR: DataOpRecordColumnFromGraphFn for in-column {} is None!".format(op_rec_column)
             out_graph_fn_column = op_rec_column.out_graph_fn_column
             # Make sure the number of returned ops matches the number of op-records in the next column.
             assert len(ops) == len(out_graph_fn_column.op_records), \
@@ -640,46 +665,140 @@ class GraphBuilder(Specifiable):
             return len(tf.get_default_graph().as_graph_def().node)
         return 0
 
-    def sanity_check_build(self, component=None):
-        # TODO: Obsolete method?
+    def sanity_check_build(self, still_building=False):
         """
-        Checks whether all the `component`'s and sub-components's in-Sockets and graph_fns are input-complete and
-        raises detailed error messages if not. Input-completeness means that ..
-        a) all in-Sockets of a Component or
-        b) all connected incoming Sockets to a GraphFunction
-        .. have their `self.space` field defined (is not None).
+        Checks whether some of the root component's API-method output columns contain ops that are still None.
+        """
+        for api_method_rec in self.root_component.api_methods.values():
+            # Ignore `_variables` API-method for now.
+            if api_method_rec.name == "_variables":
+                continue
+            for out_op_column in api_method_rec.out_op_columns:
+                for op_rec in out_op_column.op_records:
+                    if op_rec.op is None:
+                        try:
+                            self._analyze_none_op(op_rec)
+                        except RLGraphBuildError as e:
+                            if still_building:
+                                print("Found problem in build process (causing a build-deadlock):")
+                            raise e  # TODO: do something else with this error?
+
+    def _analyze_none_op(self, op_rec):
+        """
+        Args:
+            op_rec (DataOpRecord): The op-rec to analyze for errors (whose op property is None).
+
+        Raises:
+            RLGraphError: After the problem has been identified.
+        """
+        # Step via `previous` through the graph backwards.
+        while True:
+            previous_op_rec = op_rec.previous
+            # Hit a graph_fn. Jump to incomign column.
+            if previous_op_rec is None:
+                assert isinstance(op_rec.column, DataOpRecordColumnFromGraphFn),\
+                    "ERROR: If previous op-rec is None, column must be of type `DataOpRecordColumnFromGraphFn` " \
+                    "(but is type={})!".format(type(op_rec.column).__name__)
+                # All op-recs going into this graph_fn have actual ops.
+                # -> We know now that this graph_fn is only not called because the Component is either input-incomplete
+                # or variable-incomplete.
+                if op_rec.column.in_graph_fn_column.is_complete():
+                    if op_rec.column.component.input_complete is False:
+                        self._analyze_input_incomplete_component(op_rec.column.component)
+                    else:
+                        assert op_rec.column.component.variable_complete is False and \
+                               op_rec.column.graph_fn_name == "_graph_fn__variables", \
+                               "ERROR: Component '{}' was expected to be either input-incomplete or " \
+                               "variable-incomplete!".format(op_rec.column.component.global_scope)
+                        self._analyze_variable_incomplete_component(op_rec.column.component)
+                else:
+                    # Take as an example None op the first incoming one that's None as well.
+                    empty_in_op_recs = list(or_ for or_ in op_rec.column.in_graph_fn_column.op_records if or_.op is None)
+                    if len(empty_in_op_recs) > 0:
+                        previous_op_rec = empty_in_op_recs[0]
+                    # All op-recs have actual ops -> .
+                    else:
+                        pass  # TODO: complete logic
+            # Continue with new op-record.
+            op_rec = previous_op_rec
+
+    def _analyze_input_incomplete_component(self, component):
+        """
+        Analyzes why a component is input-incomplete and what we can further track back from it
+        (e.g. maybe there is another one before that that is also input-incomplete).
 
         Args:
-            component (Component): The Component to analyze for input-completeness.
+            component (Component): The defunct Component to analyze.
+
+        Raises:
+            RLGraphError: After the problem has been identified.
         """
-        component = component or self.root_component
+        self.investigated_input_incomplete_components.add(component)
+        # Find any input-param that has no Space defined.
+        incomplete_input_args = list(name for name, space in component.api_method_inputs.items() if space is None)
+        assert len(incomplete_input_args) > 0, "ERROR: Expected at least one input-arg to be without Space-definition!"
+        incomplete_arg = incomplete_input_args[0]
+        ## Either we are blocked by ourselves (API-method calling another one of the same Component (internal deadlock)).
+        #internal_deadlock = False
+        # Now loop through all of the Component's API-methods and look for all calls using `incomplete_arg`.
+        # If they are all coming from the same Component -> internal deadlock.
+        # If at least one is coming from another component -> follow that one via this very method.
+        # TODO: What if there is external deadlock? Two components
+        calls_using_incomplete_arg = 0
+        components_making_calls_with_incomplete_arg = set()
+        for api_method_name, api_method_rec in component.api_methods.items():
+            # Found a call with `incomplete_arg`.
+            if incomplete_arg in api_method_rec.input_names and len(api_method_rec.in_op_columns) > 0:
+                calls_using_incomplete_arg += len(api_method_rec.in_op_columns)
+                for call_column in api_method_rec.in_op_columns:
+                    for op_rec in call_column.op_records:
+                        assert op_rec.previous is not None
+                        components_making_calls_with_incomplete_arg.add(op_rec.previous.column.component)
 
-        if self.logger.level <= logging.INFO:
-            component_print_out(component)
+        must_be_complete_suggestion = \
+            "If the space for this arg is not important in creating variables for this component, try flagging the " \
+            "API-methods that use this arg via the `must_be_complete=False` flag."
 
-        # Check all the component's graph_fns for input-completeness.
-        for graph_fn in component.graph_fns:
-            if graph_fn.input_complete is False:
-                # Look for the missing in-Socket and raise an Error.
-                for in_sock_name, in_sock_record in graph_fn.input_sockets.items():
-                    if len(in_sock_record["socket"].op_records) == 0 and \
-                            in_sock_name not in component.unconnected_sockets_in_meta_graph:
-                        raise RLGraphError("in-Socket '{}' of GraphFunction '{}' of Component '{}' does not have "
-                                           "any incoming ops!".format(in_sock_name, graph_fn.name,
-                                                                      component.global_scope))
+        # What if incomplete_arg was never used in any calls?
+        # Then that's the reason, why this component is input-incomplete.
+        if calls_using_incomplete_arg == 0:
+            raise RLGraphBuildError(
+                "The call argument '{}' in Component '{}' was never used in any calls to any API-method of this "
+                "component! Thus, the component remains input-incomplete. "
+                "{}".format(incomplete_arg, component.global_scope, must_be_complete_suggestion)
+            )
+        # Only this very component uses this call arg -> "Inner deadlock".
+        elif len(components_making_calls_with_incomplete_arg) == 1 and \
+                component in components_making_calls_with_incomplete_arg:
+            raise RLGraphBuildError(
+                "Component '{}' has a circular dependency via API call arg '{}'! Only this component ever makes "
+                "calls using this arg, so it can never become input-complete. "
+                "{}".format(component.global_scope, incomplete_arg, must_be_complete_suggestion)
+            )
+        # Some other component(s) use this call arg, but they might be input-incomplete themselves.
+        else:
+            for calling_component in components_making_calls_with_incomplete_arg:
+                if calling_component is component:
+                    continue
+                # Assume that the caller is input-incomplete itself.
+                assert calling_component.input_complete is False
+                # Continue investigating why this one is input incomplete.
 
-        # Check component's sub-components for input-completeness (recursively).
-        for sub_component in component.sub_components.values():  # type: Component
+    def _analyze_variable_incomplete_component(self, component):
+        """
+        Analyzes why a component is variable-incomplete (one of its children is not input-complete) and keeps tracking
+        the root cause for this problem.
+
+        Args:
+            component (Component): The defunct Component to analyze.
+
+        Raises:
+            RLGraphError: After the problem has been identified.
+        """
+        # Find the sub-component that's input-incomplete and further track that one.
+        for sub_component in component.get_all_sub_components():
             if sub_component.input_complete is False:
-                # Look for the missing Socket and raise an Error.
-                for in_sock in sub_component.input_sockets:
-                    if in_sock.space is None:
-                        raise RLGraphError("Component '{}' is not input-complete. In-Socket '{}' does not " \
-                                           "have any incoming connections.".
-                                           format(sub_component.global_scope, in_sock.name))
-
-            # Recursively call this method on all the sub-component's sub-components.
-            self.sanity_check_build(sub_component)
+                self._analyze_input_incomplete_component(sub_component)
 
     def get_execution_inputs(self, *api_method_calls):
         """
@@ -713,10 +832,18 @@ class GraphBuilder(Specifiable):
             if len(self.api[api_method_call][1]) > 0 and self.api[api_method_call][1][0].kwarg is not None:
                 fetch_dict[api_method_call] = {op_rec.kwarg: op_rec.op for op_rec in self.api[api_method_call][1] if
                                                return_ops is None or op_rec.kwarg in return_ops}
+                if return_ops is not None:
+                    assert all(op in fetch_dict[api_method_call] for op in return_ops),\
+                        "ERROR: Not all wanted return_ops ({}) are returned by API-method `api_method_call`!".format(
+                        return_ops)
             # API returns a tuple.
             else:
                 fetch_dict[api_method_call] = [op_rec.op for i, op_rec in enumerate(self.api[api_method_call][1]) if
                                                return_ops is None or i in return_ops]
+                if return_ops is not None:
+                    assert len(fetch_dict[api_method_call]) == len(return_ops),\
+                        "ERROR: Not all wanted return_ops ({}) are returned by API-method `api_method_call`!".format(
+                        return_ops)
 
             for i, param in enumerate(params):
                 # TODO: What if len(params) < len(self.api[api_method][0])? Need to handle default API-method params also for the root-component (this one).
@@ -725,13 +852,14 @@ class GraphBuilder(Specifiable):
                                        "{}.".format(api_method_call, len(self.api[api_method_call][0]), len(params)))
 
                 placeholder = self.api[api_method_call][0][i].op  # 0=input op-recs; i=ith input op-rec
-                if isinstance(placeholder, DataOpTuple):
-                    for ph, p in zip(placeholder, param):
-                        feed_dict[ph] = p
+                if isinstance(placeholder, ContainerDataOp):
+                    flat_placeholders = flatten_op(placeholder)
+                    for flat_key, value in flatten_op(param).items():
+                        feed_dict[flat_placeholders[flat_key]] = value
                 # Special case: Get the default argument for this arg.
                 # TODO: Support API-method's kwargs here as well (mostly useful for test.test).
                 #elif param is None:
-                #    feed_dict[placeholder] = self.root_component.api_methods[api_method].default_values[i]
+                #    feed_dict[placeholder] = self.root_component.api_methods[api_method_call].default_values[i]
                 else:
                     feed_dict[placeholder] = param
 
@@ -765,10 +893,106 @@ class GraphBuilder(Specifiable):
             else:
                 return self.root_component.api_fn_by_name[api_method]()
 
+    def execute_define_by_run_graph_fn(self, component, graph_fn, options, *args, **kwargs):
+        """
+        Executes a graph_fn in define by run mode.
+
+        Args:
+            graph_fn (callable): Graph function to execute.
+            options (dict): Execution options.
+        Returns:
+            any: Results of executing this graph-fn.
+        """
+        flatten_ops = options.pop("flatten_ops", False)
+        split_ops = options.pop("split_ops", False)
+        add_auto_key_as_first_param = options.pop("add_auto_key_as_first_param", False)
+        # print("executing graph fn", graph_fn)
+        # print("raw args = ", args)
+        # print("raw kwargs = ", kwargs)
+
+        # No container arg handling.
+        if not flatten_ops:
+            return graph_fn(component, *args, **kwargs)
+        else:
+            # Flatten and identify containers for potential splits.
+            flattened_args = []
+
+            # Was there actually any flattening
+            args_actually_flattened = False
+            for arg in args:
+                if isinstance(arg, (Dict, dict, tuple)) or isinstance(arg, Dict) or isinstance(arg, tuple):
+                    flattened_args.append(define_by_run_flatten(arg))
+                    args_actually_flattened = True
+                else:
+                    flattened_args.append(arg)
+
+            flattened_kwargs = {}
+            if len(kwargs) > 0:
+                for key, arg in kwargs.items():
+                    if isinstance(arg, dict) or isinstance(arg, Dict) or isinstance(arg, tuple):
+                        flattened_kwargs[key] = define_by_run_flatten(arg)
+                        args_actually_flattened = True
+                    else:
+                        flattened_kwargs[key] = arg
+
+            # If splitting args, split then iterate and merge. Only split if some args were actually flattened.
+            if args_actually_flattened:
+                split_args_and_kwargs = define_by_run_split_args(add_auto_key_as_first_param,
+                                                                 *flattened_args, **flattened_kwargs)
+
+                # Idea: Unwrap light flattening by iterating over flattened args and reading out "" where possible
+                if split_ops and isinstance(split_args_and_kwargs, OrderedDict):
+                    # Args were actually split.
+                    ops = {}
+                    num_return_values = -1
+                    for key, params in split_args_and_kwargs.items():
+                        params_args = [p for p in params if not isinstance(p, tuple)]
+                        params_kwargs = {p[0]: p[1] for p in params if isinstance(p, tuple)}
+                        ops[key] = force_tuple(graph_fn(component, *params_args, **params_kwargs))
+                        num_return_values = len(ops[key])
+
+                    # Un-split the results dict into a tuple of `num_return_values` slots.
+                    un_split_ops = []
+                    for i in range(num_return_values):
+                        dict_with_singles = OrderedDict()
+                        for key in split_args_and_kwargs.keys():
+                            dict_with_singles[key] = ops[key][i]
+                        un_split_ops.append(dict_with_singles)
+
+                    flattened_ret = tuple(un_split_ops)
+                else:
+                    split_args = split_args_and_kwargs[0]
+                    split_kwargs = split_args_and_kwargs[1]
+                    # Args did not contain deep nested structure so
+                    flattened_ret = graph_fn(component, *split_args, **split_kwargs)
+
+                # If result is a raw tensor, return as is.
+                if get_backend() == "pytorch":
+                    if isinstance(flattened_ret, torch.Tensor):
+                        return flattened_ret
+
+                unflattened_ret = []
+                for i, op in enumerate(flattened_ret):
+                    # Try to re-nest ordered-dict it.
+                    if isinstance(op, OrderedDict):
+                        unflattened_ret.append(define_by_run_unflatten(op))
+                    # All others are left as-is.
+                    else:
+                        unflattened_ret.append(op)
+
+                # Return unflattened results.
+                return unflattened_ret
+            else:
+                # Just pass in args and kwargs because not actually flattened, with or without default key.
+                if add_auto_key_as_first_param:
+                    return graph_fn(component, "", *args, **kwargs)
+                else:
+                    return graph_fn(component, *args, **kwargs)
+
     def build_define_by_run_graph(self, meta_graph, input_spaces, available_devices,
                                   device_strategy="default", default_device=None, device_map=None):
         """
-        Builds a graph for eager execution. This primarily consists of creating variables through
+        Builds a graph for eager or define by run execution. This primarily consists of creating variables through
         the component hierarchy by pushing the input spaces  through the graph.
 
           Args:
@@ -794,6 +1018,7 @@ class GraphBuilder(Specifiable):
         self.device_strategy = device_strategy
         self.default_device = default_device
         self.device_map = device_map or {}
+        self.phase = "building"
 
         # TODO device strategy in pytorch?
         # Build full registry of callable methods on root component.
@@ -860,7 +1085,7 @@ class GraphBuilder(Specifiable):
                             format(op_rec, next_op_rec, next_op_rec, next_op_rec.previous)
                         # If not last op in this API-method -> continue.
                         if next_op_rec.is_terminal_op is False:
-                            assert next_op_rec.op is None or is_constant(next_op_rec.op)
+                            assert next_op_rec.op is None or is_constant(next_op_rec.op) or next_op_rec.op is op_rec.op
                             self.op_records_to_process.add(next_op_rec)
                         # Push op and Space into next op-record.
                         next_op_rec.op = op_rec.op
@@ -876,8 +1101,8 @@ class GraphBuilder(Specifiable):
                             if component.api_method_inputs[param_name] is None or \
                                     component.api_method_inputs[param_name] == "flex":
                                 component.api_method_inputs[param_name] = next_op_rec.space
-                            # Sanity check, whether Spaces are equivalent.
-                            else:
+                            # For non-space agnostic Components: Sanity check, whether Spaces are equivalent.
+                            elif component.space_agnostic is False:
                                 generic_space = check_space_equivalence(
                                     component.api_method_inputs[param_name], next_op_rec.space
                                 )
@@ -889,9 +1114,7 @@ class GraphBuilder(Specifiable):
                                                        component.api_method_inputs[param_name])
                                     )
                                 # Overwrite both entries with the more generic Space.
-                                else:
-                                    next_op_rec.space = component.api_method_inputs[param_name] = \
-                                        generic_space
+                                next_op_rec.space = component.api_method_inputs[param_name] = generic_space
 
                         # Did we enter a new Component? If yes, check input-completeness and
                         # - If op_rec.column is None -> We are at the very beginning of the graph (op_rec.op is a
@@ -927,14 +1150,10 @@ class GraphBuilder(Specifiable):
             # Sanity check, whether we are stuck.
             new_op_records_list = sorted(self.op_records_to_process, key=lambda rec: rec.id)
             if op_records_list == new_op_records_list:
-                # Ok for some
-                if loop_counter > 10000:
-                    #print()
-                    raise RLGraphError(
-                        "Build procedure is deadlocked. Most likely, you are having a circularly dependent Component "
-                        "in your meta-graph. The following Components are still input-incomplete:\n{}".
-                        format(non_complete_components)
-                    )
+                # Probably deadlocked. Do a premature sanity check to report possible problems.
+                if loop_counter > self.max_build_iterations:
+                    self.sanity_check_build(still_building=True)
+                    return
 
             op_records_list = new_op_records_list
 
@@ -965,7 +1184,7 @@ class GraphBuilder(Specifiable):
                         )}
                         var_space = Dict(var_spaces)
                         op_rec.space = var_space
-                        op_rec.op = self.get_placeholder("api-", space=var_space, component=self.root_component)
+                        op_rec.op = self.get_placeholder("api-var-input", space=var_space, component=self.root_component)
 
             loop_counter += 1
         return loop_counter
