@@ -80,6 +80,7 @@ class TensorFlowExecutor(GraphExecutor):
         # devices depending on the device strategy and we hence keep an instance here to be able
         # to request special device init ops.
         self.optimizer = None
+        self.vf_optimizer = None
 
         self.graph_default_context = None
         self.local_device_protos = device_lib.list_local_devices()
@@ -157,7 +158,7 @@ class TensorFlowExecutor(GraphExecutor):
                 # if self.default_device not in self.available_devices:
                 #    raise RLGraphError("Provided `default_device` ('{}') is not in `available_devices` ({})".
                 #                       format(self.default_device, self.available_devices))
-            self.device_map = dict()
+            self.device_map = {}
             # Clean up device map so it only contains devices that are actually available (otherwise,
             # use the default device).
             for component_name, device in self.execution_spec["device_map"].items():
@@ -171,8 +172,8 @@ class TensorFlowExecutor(GraphExecutor):
     def build(self, root_components, input_spaces, optimizer=None, build_options=None, batch_size=32):
         # Use perf_counter for short tasks.
         start = time.perf_counter()
-        # 0. Init phase: Component construction and nesting (child/parent Components).
-        # Components can still be modified and re-arranged after this.
+
+        # Check graph setup and construct the static graph object.
         self.init_execution()
         self.setup_graph()
 
@@ -184,7 +185,10 @@ class TensorFlowExecutor(GraphExecutor):
         build_times = []
 
         for component in root_components:
-            self._build_device_strategy(component, optimizer, batch_size=batch_size)
+            # Sanity-check the component tree (from root all the way down).
+            self.sanity_check_component_tree(root_component=component)
+
+            self._build_device_strategy(component, optimizer, batch_size=batch_size, extra_build_args=build_options)
             start = time.perf_counter()
             meta_graph = self.meta_graph_builder.build(component, input_spaces)
             meta_build_times.append(time.perf_counter() - start)
@@ -446,6 +450,9 @@ class TensorFlowExecutor(GraphExecutor):
         # TODO let graph builder do this
         if self.optimizer is not None:
             var_list.extend(self.optimizer.get_optimizer_variables())
+            # If the VF has a separate optimizer (non-shared network), we need to fetch its vars here as well.
+            if self.vf_optimizer is not None:
+                var_list.extend(self.vf_optimizer.get_optimizer_variables())
 
         if self.execution_mode == "single":
             self.init_op = tf.variables_initializer(var_list=var_list)
@@ -633,7 +640,7 @@ class TensorFlowExecutor(GraphExecutor):
         # Close the tf.Session.
         self.monitored_session.close()
 
-    def _build_device_strategy(self, root_component, root_optimizer, batch_size):
+    def _build_device_strategy(self, root_component, root_optimizer, batch_size, extra_build_args=None):
         """
         When using multiple GPUs or other special devices, additional graph components
         may be required to split up incoming data, load it to device memories, and aggregate
@@ -650,8 +657,12 @@ class TensorFlowExecutor(GraphExecutor):
             root_component (Component): The root Component (will be used to create towers via `Component.copy()`).
             root_optimizer (Optimizer): The Optimizer object of the root Component.
             batch_size (int): The batch size that needs to be split between the different GPUs.
+            extra_build_args (Optional[dict]): Extra build elements to pass.
         """
         self.optimizer = root_optimizer
+        # Save separate vf optimizer if necessary.
+        if extra_build_args is not None and "vf_optimizer" in extra_build_args:
+            self.vf_optimizer = extra_build_args["vf_optimizer"]
 
         if self.device_strategy == "multi_gpu_sync":
             assert self.num_gpus > 1, "ERROR: MultiGpuSync strategy needs more than one GPU available but" \
