@@ -21,13 +21,17 @@ from rlgraph.tests.agent_test import AgentTest
 from rlgraph.tests.component_test import ComponentTest
 from rlgraph.tests.test_util import recursive_assert_almost_equal
 from rlgraph.tests.dummy_components import *
+from rlgraph.tests.dummy_components_with_sub_components import *
 
 
 __all__ = [
     "recursive_assert_almost_equal",
-    "ComponentTest",
+    "AgentTest", "ComponentTest",
+    # Simple dummies.
     "Dummy0To1", "Dummy1To1", "Dummy1To2", "Dummy2To1", "Dummy2GraphFns1To1",
     "DummyWithVar", "SimpleDummyWithVar",
+    "FlattenSplitDummy", "NoFlattenNoSplitDummy",  "OnlyFlattenDummy",
+    # Dummies with sub-components.
     "DummyWithSubComponents", "DummyCallingSubComponentsAPIFromWithinGraphFn",
-    "FlattenSplitDummy", "NoFlattenNoSplitDummy",  "OnlyFlattenDummy"
+    "DummyNNWithDictInput", "DummyProducingInputIncompleteBuild"
 ]
