@@ -109,7 +109,7 @@ class DataOpRecordColumn(object):
                     # Do constant value assignment here.
                     elif args[i] is not None:
                         op = args[i]
-                        if is_constant(op):
+                        if is_constant(op) and not isinstance(op, np.ndarray):
                             op = np.array(op, dtype=dtype(type(op), "np"))
                         op_rec.op = op
                         op_rec.space = get_space_from_op(op)
@@ -190,6 +190,7 @@ class DataOpRecordColumnIntoGraphFn(DataOpRecordColumn):
     """
     def __init__(self, component, graph_fn, flatten_ops=False,
                  split_ops=False, add_auto_key_as_first_param=False, args=None, kwargs=None):
+        # TODO: need to pass in input-arg name information so we can switch to kwargs if some default arg not given.
         super(DataOpRecordColumnIntoGraphFn, self).__init__(
             component=component, args=args, kwargs=kwargs
         )
@@ -289,7 +290,7 @@ class DataOpRecordColumnIntoGraphFn(DataOpRecordColumn):
                 for op in ops:
                     params.append(op[key] if key in op else op[""])
                 # Add kwarg_ops
-                for kwarg_key, kwarg_op in kwarg_ops:
+                for kwarg_key, kwarg_op in kwarg_ops.items():
                     params.append(tuple([
                         kwarg_key,
                         kwarg_ops[kwarg_key][key] if key in kwarg_ops[kwarg_key] else kwarg_ops[kwarg_key][""]
@@ -300,7 +301,7 @@ class DataOpRecordColumnIntoGraphFn(DataOpRecordColumn):
         # We don't have any container ops: No splitting possible. Return args and kwargs as is.
         else:
             return tuple(([""] if self.add_auto_key_as_first_param is True else []) + [op[""] for op in ops]),\
-                   {key: value[""] for key, value in kwarg_ops}
+                   {key: value[""] for key, value in kwarg_ops.items()}
 
     @staticmethod
     def unflatten_output_ops(*ops):
@@ -468,7 +469,7 @@ class GraphFnRecord(object):
 
 def get_call_param_name(op_rec):
     api_method_rec = op_rec.column.api_method_rec  # type: APIMethodRecord
-    pos_past_normals = op_rec.position - len(api_method_rec.non_args_kwargs)
+    pos_past_normals = None if op_rec.position is None else op_rec.position - len(api_method_rec.non_args_kwargs)
 
     # There are *args in the signature.
     if api_method_rec.args_name is not None:
